#include "Mreza.h"
#include <cmath>
#include <cstdlib>

Mreza::Mreza(int bc, int bl, int ops)
{
	listaL = 0;
	listaC=poslC=0;
	brC=bc;
	brL=bl;
	opseg=ops;
	uspK=neK=0;						// brojac uspesnih i neuspesnih konekcija
	veze=0;
	Cvor *c=new Cvor(0, this);			// pravi prvi cvor
	ElemC *e=new ElemC(c);
	listaC=poslC=e;					// ubacuje inicijalni cvor u listu
	for(int i=1;i<brC;i++)				// popunjava listu cvorova
	{
		c=new Cvor(i, this);
		e=new ElemC(c);
		poslC->sled=e;
		poslC=poslC->sled;
	}

	if (brL > brC * (log(brC)/log(3)))			// uslov biranja topologije
	{
		zvedasta();
	}
	else 
	{
		prstenasta();
	}
}
Cvor& Mreza::pronadjiC(int id)
{
	ElemC *tekC=listaC;
	while(tekC->p->identitet()!=id)		//postavi tek na cvor identifikacionog broja, id
		tekC=tekC->sled;
	Cvor *a;
	a=tekC->p;
	return *a;
}
void Mreza::zvedasta()
{
	cout<<"\n topologija ZVEZDA \n";
	Cvor *c1=0, *c2=0;
	int povez=0;											// brojac do sada ostvarnih povezivanja
	for(int i=0;(i<brC-1)&&(povez<brL);i++)
	{
		c1 = &pronadjiC(i);
		for(int j=i+1;(j<brC)&&(povez<brL);j++)
		{
			c2 = &pronadjiC(j);
			c1->dodaj(c2, opseg);
			povez++;
		}
	}
}
void Mreza::prstenasta()
{
	cout<<"\n topologija PRSTENA \n";
	Cvor *c1=0, *c2=0;
	int povez=0;										// brojac do sada ostvarnih povezivanja
	while(povez<brC)									// Inicijalno povezuje sve cvorove u krug (susedi su mu id cvora +/-1)
	{
		c1 = &pronadjiC(povez % brC);
		c2 = &pronadjiC((povez+1)% brC);
		c1->dodaj(c2, opseg);
		povez++;
	}
	// Ukoliko ima preostalih linkova slobodnih dodljuje dva slucajna cvora koja do sad nisu povezani
	while(povez<brL)									
	{
		int a= rand() % brC;
		int b= rand() % brC;
		if (a!=b)
		{
			c1=&pronadjiC(a);
			c2=&pronadjiC(b);
			if(!(c1->povezan(c2)))		// ukoliko ne postoji veza izmedju cvorva c1 i c2
			{
				c1->dodaj(c2, opseg);
				povez++;
			}
		}	
	}
}
void Mreza::DodajLink(Link* l)
{
	ElemL* novi = new ElemL(l);
	novi->sled = listaL;
	listaL = novi;
}
void Mreza::zauzetost(int proc)
{
	ElemC *tekC=listaC;
	Cvor *pom=0;
	for(int i=0;i<brC;i++)
	{
		if(rand() % 100 < proc)
		{
			int f=rand()% opseg;
			tekC->p->zauzmiF(f);
		}
		tekC=tekC->sled;
	}
}
int Mreza::prohodnost(int id1,int id2)
{
	Cvor *c1=0, *c2=0, *pom=0;;
	c1 = &pronadjiC(id1);
	c2 = &pronadjiC(id2);

	Red *red=new Red();

	int *provereni=new int[brC];
	/* ispitivanje za svaku frekvenciju */
	for(int i=0; i<opseg; i++)
	{
		red->dodaj(id1);
		provereni[0]=id1;				//interni spisak cvorova koji su vec obradjeni
		for (int loc=1; loc<brC; provereni[loc++] =- 1);
		/* sve dok ima potencijalnih cvorova koji su dostupni za frekvenciju koja se ispituje */
		while(!red->prazan())
		{
			int pt=red->dohvati();
			pom=&pronadjiC(pt);
			/* da li se cvor koji se trenutno obradjuje granici sa cvorom koji je destinacija */
			if(!(pom->povezan(c2)))
			{
				int j=0;
				/* sve dok cvor koji se trenutno obradjuje ima suseda */
				while(pom->susedni(j)!=-1)
				{
					int k=0;
					bool ind=true;
					/* sused cvora nije do sad bio posecen */
					while((provereni[k] != -1) && (ind))
					{
						if(pom->susedni(j)==provereni[k])ind =false;
						k++;
					}
					if(ind)
					{
						/* ukoliko je frekvencija slobodna stavi cvor u spisak cvorova za dalju proveru */
						if(pom->prvoeraVeze(j,i))
						{
							pt=pom->susedni(j);
							red->dodaj(pt);
							provereni[k]=pt;
						}
					}
					j++;
				}
			}
			//ako je cvor destinacije sused trenutno obrajdivanom cvoru proveravamo dostupnost na trazenoj frekvenciji
			else
			{
				int j=pom->povezanNa(c2);
				if(pom->prvoeraVeze(j,i))
				{
					delete red;
					delete provereni;
					return i; // vracamo frekvenciju ukoliko je slobodna
				}
			}
		}
		// praznjenje reda priprema za prolazak ispitivanja za sledecu frekvenciju
		while(!red->prazan())
		{
			int a=red->dohvati();
		}
	}
	// nije moguce ostvariti vezu izmedju ova dva cvora ni na jednoj frekvenciji
	delete red;
	delete provereni;
	return -1;
}
int Mreza::najblizi(int id1,int id2,int f)
{
	Cvor *c1=0, *c2=0, *pom=0;;
	c1 = &pronadjiC(id1);
	c2 = &pronadjiC(id2);

	Red *red=new Red();

	int *provereni=new int[brC];							//interni spisak cvorova koji su vec obradjeni
		
	red->dodaj(id1);
	for (int loc=0; loc<brC;loc++) provereni[loc]=-1;
	provereni[0]=id1;
	/* sve dok ima potencijalnih cvorova koji su dostupni za frekvenciju koja se ispituje */
	while(!red->prazan())
	{
		int pt=red->dohvati();
		pom=&pronadjiC(pt);

		int j=0;
		/* sve dok cvor koji se trenutno obradjuje ima suseda */
		while(pom->susedni(j)!=-1)
		{
		/* da li se cvor koji se trenutno obradjuje granici sa cvorom koji je destinacija */
			if(pom->povezan(c2))
			{
				int u=pom->povezanNa(c2);
				//ako je cvor destinacije sused trenutno obrajdivanom cvoru proveravamo dostupnost na trazenoj frekvenciji
				if(pom->prvoeraVeze(u,f))
				{
					delete red;
					delete provereni;
					return pom->identitet();// vracamo id cvora koji ima direktnu vezu sa destinacijom
				}
			}
			int k=0;
			bool ind=true;
			while((provereni[k] != -1) && (ind))
			{
				if(pom->susedni(j)==provereni[k])ind =false;
				k++;
			}
			/* sused cvora nije do sad bio posecen */
			if(ind)
			{
				if(pom->prvoeraVeze(j,f))
				{
					pt=pom->susedni(j);/* ukoliko je frekvencija slobodna stavi cvor u spisak cvorova za dalju proveru */
					red->dodaj(pt);
					provereni[k]=pt;
				}
			}
			j++;
		}
	}
	delete red;
	delete provereni;
	return -1;// ne postoji cvor koji ima direktnu vezu sa destinacijom i slobodan je na frekvenciji f
}
void Mreza::zauzmi(int proc)
{
	ElemV *temp=0;
	for(int i=0;i<brC;i++)
	{
		temp=veze;
		int povez = rand() % 100;
		if(povez<proc)
		{
			while(1)
			{
				povez=rand() % brC;
				if(povez!=i)
				{
					break;
				}
			}
			bool ind=true;
			// proveravamo da li zeljeni par za povezivanje vec poseduje ostvarenu vezu
			while(temp && ind)
			{
				if(temp->c1==i || temp->c1==povez)
				{
					if(temp->c2==i || temp->c2==povez)
					{
						ind=false;
					}
				}
				temp=temp->sled;
			}
			if(ind)
			{
				// odma proveravamo da li je put izmedju cvorova slobodan na nekoj frekfenciji
				int mogucaf=prohodnost(i, povez);
				if(mogucaf!=-1)
				{
					uspK++;
					konektuj(i,povez,mogucaf);
				}
				else
				{
					neK++;
				}
			}
		}
	}
}
void Mreza::DodajVezu(Red *r, int a, int b, int f)
{
	ElemV *novi = new ElemV(r);
	novi->postavi(a,b,f);
	novi->sled = veze;
	veze = novi;
}
void Mreza::konektuj(int c1, int c2, int f)
{
	Cvor *pom1=0, *pom2=0;
	Red *r=new Red;
	int *put=new int[brC];
	for(int i=0; i<brC; i++)
	{
		put[i]=-1;
	}
	int p=c2;
	int p2=-1;
	int count=0;
	//pravi putanju izmedju cvorova c1 i c2
	while(p2!=c1)
	{
		p2=najblizi(c1,p,f);
		if(p2!=-1)
		{
			pom1=&pronadjiC(p);
			pom1->zauzmiF(f);
			put[count]=p2;
			p=p2;
			count++;
		}
	}
	count--;

	// zauzimanje frekfencije na linkovima izmedju cvorova od polaznog cvora do destinacionog cvora
	pom1=&pronadjiC(c1);
	pom1->zauzmiF(f);
	int i=0;
	while(i<count)
	{
		pom1=&pronadjiC(put[i]);
		pom1->zauzmiF(f);
		i++;
	}
	while(count>0)
	{
		r->dodaj(put[count-1]);
		count--;
	}
	DodajVezu(r,c1,c2,f);
}
void Mreza::PrekidVeza(int proc)
{
	ElemV *temp=veze;
	while(temp)
	{
		int a=rand() % 100;
		if(a<=proc)
		{
			if(temp==veze)
			{
				if(veze->sled)
				{
					veze=veze->sled;
					oslobodi(temp);
					temp=veze;
				}
				else
				{
					temp=veze=0;
				}
			}
			else
			{
				
				ElemV *pom=veze;
				while(pom->sled!=temp)
				{
					pom=pom->sled;
				}
				pom->sled=temp->sled;
				oslobodi(temp);
				temp=pom->sled;
			}
		}
		else
		{
			temp=temp->sled;
		}
	}
}
void Mreza::oslobodi(ElemV *v)
{
	Cvor *pom=0;
	pom = &pronadjiC(v->c1);
	pom->oslobodiF(v->f);
	while(!v->r->prazan())
	{
		pom = &pronadjiC(v->r->dohvati());
		pom->oslobodiF(v->f);
	}
	pom = &pronadjiC(v->c2);
	pom->oslobodiF(v->f);
	delete v;
}
void Mreza::statistika()
{
	cout<<"od ukupno "<< uspK+neK <<"uspesno je realizovano "<<uspK<<" \n";
}
Mreza::~Mreza(void)
{
	while(listaC)				// obrisi sve cvorove u mrezi
	{
		ElemC *tekC=listaC;
		listaC=listaC->sled;
		delete tekC;
	}

	while(listaL)				// obrisi sve linkove u mrezi
	{
		ElemL* temp = listaL;
		listaL=listaL->sled;
		delete temp;
	}
	while(veze)					// brise sve uspostavljene veze
	{
		ElemV *temp = veze;
		veze=veze->sled;
		delete temp;
	}
}
int main()
{
	int brC, brL, opsK;
	while(1)											// Proces unosenja broja cvorova sa zadatim uslovima
	{
		cout <<"\n Unesite broj cvorova:\n	";
		cin >> brC;
		if(brC>=3)break;else
		{
			cout << "\n\n Broj cvorova mora da bude veci od 3 \n";
		}
	}
	// Proces unosenja broja linkova
	int Lmax=0;										//	Maksimalni broj linkova u zavisnosti od broja cvorova
	for(int i=1;i<=brC;i++)Lmax+=brC-i;				//	Racunanje maksimalnog broja linkova
	cout <<"\n Unesite broj linkova:\n	";
	cin >> brL;
	//	Minimalni broj linkova za bilo koji broj zadatih cvorova jeste broj cvorova
	if(brC>brL)
	{
		cout << "Broj linkova je povecan na minimalnih :"<<brC<<"\n";
		brL=brC;
	}
	if(brL>Lmax)
	{
		cout << "\n Broj linkova je smanjen na maksimalnih  :"<<Lmax<<"\n ";
		brL=Lmax;
	}
	// Unosenje broja opsega na linkovima
	cout << "\n Unesite koliko svaki od linkova ima razlicitih frekfencija na raspolaganju: \n	";
	cin >> opsK;
	if(opsK<=0)
	{
		cout << "\n Broj kanala je automatski povecan na 1\n";
		opsK=1;
	}
	Mreza m(brC,brL,opsK);
	int opcija=1;
	while(opcija)
	{
		cout <<"\n da li zelite da izvrsite simulaciju za zadatak 2 (2), zadatak 3 (3) ili da napusite progam(0) \n";
		cin >> opcija;
		if(opcija==2)
		{
			int proc1=0;
			int ukupno=0,uspesno=0;
			cout <<"\n unesite verovatnocu zauzetosti cvora u %(primer 21, 30, 85....) u 1-om vremenskom intervalu :\n";
			cin >> proc1;
			m.zauzetost(proc1);
			for(int i=0;i<brC-1;i++)
			{
				for(int j=i+1;j<brC;j++)
			{
			ukupno++;
			if(m.prohodnost(i,j)!=-1)uspesno++;
			}
		}
		cout <<"\n od maksimalnih :"<<ukupno<<"zahteva za komunikaciju medju cvorovima, moguce je ostvariti :"<<uspesno<<" konekcija medju razlicitim cvorovima za slucaj kada je verovatnoca zauzetosti svakog od covorova:"<<proc1<<"%. \n";
		int c1=0, c2=0;
		cout <<"\n unesite scvor 1 \n";
		cin >>c1;
		cout <<"\n unesite scvor 2 \n";
		cin >>c2;
		if(c1==c2)cout<<"\n cvorovi 1 i 2 moraju biti razliciti \n";
		else
		if(m.prohodnost(c1,c2)!=-1)
		{
			cout<<"\n uspesno konektovanje izmedju"<<c1<<" i "<<c2<<" \n";
		}
		else
		{
			cout <<"\n nije moguce uspostaviti vezu izmedju ova dva cvora\n";
		}
		}
		if(opcija==3)
		{
			int proc1=0;
			int proc2=0;
			int vreme;
			int ukupno=0,uspesno=0;
			cout <<"\n unesite verovatnocu uspostavljanja konekcije cvora u %(primer 21, 30, 85....) ka slucajno odabranom cvoru u 1-om vremenskom intervalu :\n";
			cin >> proc1;
			cout <<"\n unestie verovatnocu gasenja svake od ostvarenih veza\n";
			cin >> proc2;
			cout <<"\n za koliko vremenskih intervala zelite da se simulacija izvrsava? \n";
			cin >> vreme;
			for(int i=0;i<vreme;i++)
			{
				m.zauzmi(proc1);
				m.PrekidVeza(proc2);
			}
			m.statistika();
		}
		opcija=0;
	}
}
